--[[

States:
-1	Display error (expected: errormsg)
0	Temp main menu (enter state)
1	The editor (will expect things to have been loaded!)
2	Syntax highlighting test
3	Scripting editor
4	Some XML testing
5	Filesystem testing
6	Simple listing of all files in the levels folder, and load a level from here
7	Display all sprites from sprites.png where you can get the number of the sprite you're hovering over
8	Just save by going to this state and typing in a name
9	Dialog test, and right click menu test
10	List of scripts, and enter one to load

Debug keys:
F12: change state
F9: display a test dialog to see how it looks
' (in editor): display all tilesets and colors in the console
\ (in editor): tileset creator/compiler
/ (in editor): instead of displaying entities as themselves, display them as an E-block

For the credits later:
Ved is made by Dav999
Some of the graphics (like the buttons) were made by Doormat
Special thanks to:
TurtleP (for the line of code that made tiles not blurry when scaled)
and of course Terry for making VVVVVV

]]

function love.load()
	require("lang/English")
	require("const")
	require("func")
	require("roomfunc")
	require("vvvvvvxml")
	require("dialog")

	if love.window == nil then
		-- Our version of löve is too old, display error message.
		print("Outdated love2d version!")
		errormsg = L.OUTDATEDLOVE
		state = -1
	else
		cons("love.load() reached")
		
		-- So that zooming in won't make tiles blurry, thanks TurtleP
		love.graphics.setDefaultFilter("nearest", "nearest")
		
		state = 0; oldstate = "none"
		debug = 0
		input = ""; takinginput = false
		__ = "_"
		multiinput = {}; currentmultiinput = 0
		
		mousepressed = false -- for some things
		
		-- Dialogs. To not have an active window, set DIAwindowani to 16. To show one, just call newdialog(message, title, showbar, canclose). -15 ~ -1 opening, 0 is normal, 1 ~ 15 closing
		dialog.init()
		nodialog = true
		
		tilesets = {}
		
		-- Load a couple of images
		cursorimg = {}
		cursorimg[0] = love.graphics.newImage("cursor/cursor.png")
		cursorimg[1] = love.graphics.newImage("cursor/cursor1.png")
		cursorimg[2] = love.graphics.newImage("cursor/cursor2.png")
		cursorimg[3] = love.graphics.newImage("cursor/cursor3.png")
		cursorimg[4] = love.graphics.newImage("cursor/cursor4.png")
		cursorimg[5] = love.graphics.newImage("cursor/entity.png")
		
		selectedtoolborder = love.graphics.newImage("selectedtool.png")
		unselectedtoolborder = love.graphics.newImage("unselectedtool.png")
		
		savebtn = love.graphics.newImage("save.png")
		loadbtn = love.graphics.newImage("load.png")
		newbtn = love.graphics.newImage("new.png")
		
		bggrid = love.graphics.newImage("bggrid.png")
		
		toolimg = {}
		for t = 1, 17 do
			toolimg[t] = love.graphics.newImage("tools/" .. t .. ".png")
		end
		
		scrollup = love.graphics.newImage("scrollup.png")
		scrolldn = love.graphics.newImage("scrolldn.png")
		
		-- Load the default tilesets
		loadtileset("tiles.png")
		loadtileset("tiles2.png")
		-- loadtileset("tiles3.png") oops oversight
		
		-- Load the spriteset
		-- loadsprites("sprites.png", 64)
		loadsprites("sprites.png", 32)
		
		-- eeeeeeeeee
		love.keyboard.setKeyRepeat(true)
		
		if love.system.getOS() == "OS X" then
			-- Cmd
			ctrl = "gui"
			dirsep = "/"
			require("filefunc_mac")
		else
			-- Ctrl
			ctrl = "ctrl"
			dirsep = "\\"
			require("filefunc_win")
		end
	end
end

function love.draw()
	if state == -1 then
		love.graphics.printf(L.FATALERROR .. anythingbutnil(errormsg) .. "\n\n" .. L.FATALEND, 10, 10, love.graphics.getWidth()-20, "left")
	elseif state == 0 then
		love.graphics.print("Placeholder main menu. Enter state: " .. input .. __, 10, 10)
		startinputonce()
	elseif state == 1 then
		-- Ok first of all are we clicking?
		if nodialog and love.mouse.isDown("l") and mouseon(64, 0, 639, 480) then
			atx = math.floor((love.mouse.getX()-64) / 16)
			aty = math.floor((love.mouse.getY()) / 16)
			
			if tilespicker then
				if tilescreator then
					if not mousepressed then
						if creatorstep == 1 then
							cons("Wall tile " .. creatorsubstep .. ": " .. (aty*40)+(atx+1)-1)
							
							tilesetblocks.creator.colors.creator.blocks[creatorsubstep] = (aty*40)+(atx+1)-1
							cb[creatorsubstep] = (aty*40)+(atx+1)-1
						elseif creatorstep == 2 then
							cons("Background tile " .. creatorsubstep .. ": " .. (aty*40)+(atx+1)-1)
							
							tilesetblocks.creator.colors.creator.background[creatorsubstep] = (aty*40)+(atx+1)-1
							ca[creatorsubstep] = (aty*40)+(atx+1)-1
						elseif creatorstep == 3 then
							cons("Spike tile " .. creatorsubstep .. ": " .. (aty*40)+(atx+1)-1)
							
							tilesetblocks.creator.colors.creator.spikes[creatorsubstep] = (aty*40)+(atx+1)-1
							cs[creatorsubstep] = (aty*40)+(atx+1)-1
						end
						
						if creatorsubstep == 30 then
							creatorstep = creatorstep + 1
							cons("Next: step " .. creatorstep)
							selectedtool = creatorstep
							
							creatorsubstep = 1
						else
							creatorsubstep = creatorsubstep + 1
						end
						
						if creatorstep == 4 then
							spitoutarrays()
							cons("*** ARRAYS COPIED TO CLIPBOARD ***")
							tilescreator = false
						end
						
						mousepressed = true
					end
				else
					cons("Tile selected: " .. (aty*40)+(atx+1)-1)
				
					selectedtile = (aty*40)+(atx+1)-1
				end
			else
				cons("Tile clicked: " .. atx .. " " .. aty .. ", set to " .. selectedtile)
				
				roomdata[roomy][roomx][(aty*40)+(atx+1)] = selectedtile
			end
		elseif nodialog and love.mouse.isDown("r") and mouseon(64, 0, 639, 480) then
			atx = math.floor((love.mouse.getX()-64) / 16)
			aty = math.floor((love.mouse.getY()) / 16)
			
			cons("Tile cleared: " .. atx .. " " .. aty)
			
			roomdata[roomy][roomx][(aty*40)+(atx+1)] = 0
		end
		
		if tilespicker then
			displaytilespicker(64, 0, tilesetnames[usedtilesets[levelmetadata[(roomy)*20 + (roomx+1)].tileset]])
		else
			love.graphics.draw(bggrid, 64, 0)
			--[[
			cons("Using room " .. (roomy)*20 + (roomx+1))
			for k,v in pairs(levelmetadata[(roomy)*20 + (roomx+1)]) do
				cons(k .. "->" .. v)
			end
			]]
			if love.keyboard.isDown("n") then
				love.graphics.setFont(tinynumbers)
			end
			displayroom(64, 0, roomdata[roomy][roomx], levelmetadata[(roomy)*20 + (roomx+1)])
			love.graphics.setFont(font8)
			displayentities(64, 0, roomx, roomy)
			
			-- Does this room have a name, perchance?
			--if levelmetadata[
		end
		
		-- Now display the cursor. If it's on the level
		if nodialog and mouseon(64, 0, 639, 480) then
			cursorx = math.floor((love.mouse.getX()-64) / 16)
			cursory = math.floor((love.mouse.getY()) / 16)
			
			love.graphics.draw(cursorimg[0], (cursorx*16)+64, (cursory*16))
		else
			cursorx = "--"
			cursory = "--"
		end
		
		-- We also want the tools on the left. But it's a scrollable area.
		love.graphics.setScissor(16, 16, 32, love.graphics.getHeight()-32)
		
		for t = 1, 17 do
			-- love.graphics.rectangle("fill", 16, (16+(48*(t-1)))+lefttoolscroll, 32, 32)
			-- Are we hovering over it? Or maybe even clicking it?
			if not nodialog or ((mouseon(16, 0, 32, 16)) or (mouseon(16, love.graphics.getHeight()-16, 32, 16)) or (not mouseon(16, (16+(48*(t-1)))+lefttoolscroll, 32, 32))) and selectedtool ~= t then
				love.graphics.setColor(255,255,255,128)
			end
			
			if nodialog and love.mouse.isDown("l") and mouseon(16, (16+(48*(t-1)))+lefttoolscroll, 32, 32) and not mouseon(16, 0, 32, 16) and not mouseon(16, love.graphics.getHeight()-16, 32, 16) then
				selectedtool = t
			end
			
			if selectedtool == t then
				love.graphics.draw(selectedtoolborder,  16, (16+(48*(t-1)))+lefttoolscroll)
			else
				love.graphics.draw(unselectedtoolborder,  16, (16+(48*(t-1)))+lefttoolscroll)
			end
			
			love.graphics.draw(toolimg[t], 16+2, (16+2+(48*(t-1)))+lefttoolscroll)
			love.graphics.setColor(255,255,255,255)
			
			if nodialog and ((not mouseon(16, 0, 32, 16)) and not (mouseon(16, love.graphics.getHeight()-16, 32, 16)) and (mouseon(16, (16+(48*(t-1)))+lefttoolscroll, 32, 32))) then
				-- Ugh this code but we're hovering over it. So display a tooltip, but don't get snipped away by the scissors.
				love.graphics.setScissor()
				love.graphics.setColor(128,128,128,192)
				love.graphics.rectangle("fill", love.mouse.getX()+15, love.mouse.getY()-10, string.len(toolnames[t])*8, 8)
				love.graphics.setColor(255,255,255,255)
				love.graphics.print(toolnames[t], love.mouse.getX()+16, love.mouse.getY()-8)
				love.graphics.setScissor(16, 16, 32, love.graphics.getHeight()-32)
			end
		end
		
		-- We're done with that now.
		love.graphics.setScissor()
		
		-- But we still want to be able to scroll!
		hoverdraw(scrollup, 16, 0, 32, 16)
		hoverdraw(scrolldn, 16, love.graphics.getHeight()-16, 32, 16)
		
		-- Are we clicking them?
		if nodialog and love.mouse.isDown("l") and mouseon(16, 0, 32, 16) then
			lefttoolscroll = lefttoolscroll + 4
			lefttoolscrollbounds()
		elseif nodialog and love.mouse.isDown("l") and mouseon(16, love.graphics.getHeight()-16, 32, 16) then
			lefttoolscroll = lefttoolscroll - 4
			lefttoolscrollbounds()
		end
		
		-- Now stuff on the right.
		hoverdraw(newbtn, love.graphics.getWidth()-96, 0, 32, 32, 2)
		hoverdraw(loadbtn, love.graphics.getWidth()-64, 0, 32, 32, 2)
		hoverdraw(savebtn, love.graphics.getWidth()-32, 0, 32, 32, 2)
		
		-- Well make them actually do something!
		if nodialog and love.mouse.isDown("l") and mouseon(love.graphics.getWidth()-96, 0, 32, 32) then
			-- New
			dialog.new("Making a new level isn't supported yet!", "", 1, 1, 0)
		elseif nodialog and love.mouse.isDown("l") and mouseon(love.graphics.getWidth()-64, 0, 32, 32) then
			-- Load. But first ask them if they want to save (make this save/don't save/cancel later, yes/no for now)
			dialog.new("Are you sure you want to load a level? You will lose any unsaved content.\n\n(dialog will be save/don't save/cancel later)", "", 1, 3, 3)
		elseif nodialog and love.mouse.isDown("l") and mouseon(love.graphics.getWidth()-32, 0, 32, 32) then
			-- Save
			tostate(8)
		end
		
		-- And coordinates.
		love.graphics.print("(" .. (roomx+1) .. "," .. (roomy+1) .. ")", love.graphics.getWidth()-56, love.graphics.getHeight()-16)
		
		-- But if we're in the tiles picker instead display the tile we're hovering on!
		if tilespicker then
			if (cursorx ~= "--") and (cursory ~= "--") then
				love.graphics.print(L.TILE .. " " .. (cursory*40)+(cursorx+1)-1, love.graphics.getWidth()-72, love.graphics.getHeight()-8)
			else
				love.graphics.print(L.TILE .. " ----", love.graphics.getWidth()-72, love.graphics.getHeight()-8)
			end
		else
			love.graphics.print("(" .. cursorx .. "," .. cursory .. ")", love.graphics.getWidth()-56, love.graphics.getHeight()-8)
		end
		
		-- Also display a smaller tiles picker for semi-undirect mode
		love.graphics.rectangle("fill", love.graphics.getWidth()-(7*16)-1, love.graphics.getHeight()-(6*16)-16-24-12-1, (6*16)+2, (5*16)+2)
		love.graphics.setColor(0,0,0,255)
		love.graphics.rectangle("fill", love.graphics.getWidth()-(7*16), love.graphics.getHeight()-(6*16)-16-24-12, (6*16), (5*16))
		love.graphics.setColor(255,255,255,255)
		displaysmalltilespicker(love.graphics.getWidth()-(7*16), love.graphics.getHeight()-(6*16)-16-24-12, selectedtileset, selectedcolor)
		
		-- And text below it.
		--[[
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(7*16)-1, love.graphics.getHeight()-16-32-2, (6*16), 8)
		love.graphics.print("<tiles2.png>", love.graphics.getWidth()-(7*16), love.graphics.getHeight()-16-32)
		]]
		
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(7*16)-1, love.graphics.getHeight()-16-32-2-12, (6*16), 8+4)
		love.graphics.print(" " .. selectedtileset, love.graphics.getWidth()-(7*16), love.graphics.getHeight()-16-32-12+2)
		love.graphics.print("v", love.graphics.getWidth()-(1.5*16), love.graphics.getHeight()-16-32-12+2)
		
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(7*16)-1, love.graphics.getHeight()-16-24-2-8, (6*16), 8+4)
		love.graphics.print(" " .. selectedcolor, love.graphics.getWidth()-(7*16), love.graphics.getHeight()-16-24-8+2)
		love.graphics.print("v", love.graphics.getWidth()-(1.5*16), love.graphics.getHeight()-16-24-8+2)
		
		hoverrectangle(128,128,128,128, love.graphics.getWidth()-(7*16)-1, love.graphics.getHeight()-16-16-2-4, (6*16), 8+4)
		love.graphics.print(tilespicker and L.HIDEALL or L.SHOWALL, love.graphics.getWidth()-(7*16), love.graphics.getHeight()-16-16-4+2)
		
		-- Dropdown for tileset?
		--[[
		if dropdown == 1 then
			--[ [
			love.graphics.print([ [-1-
			 Space St.
			-2-
			 Outside
			 Lab
			 Warp Zone
			 Ship] ], love.graphics.getWidth()-(7*16)-1, love.graphics.getHeight()-16-32-2-12-56)
			 ] ]
			if not nodialog then
				rightclickmenu.create({"#1", "Space St.", "#2", "Outside", "Lab", "Warp Zone", "Ship"}, "tileset")
			end
		elseif dropdown == 2 then
		
		end
		]]
		
		-- Do we want to see room metadata?
		if allowdebug and love.keyboard.isDown("f11") then
			
		end
	elseif state == 2 then
		love.graphics.print("Syntax highlighting" .. input .. __, 10, 10)
		
		syntaxhl(input, 10, 30, false, true)
		startinputonce()
	elseif state == 3 then
		love.graphics.setColor(255,255,255,255)
		love.graphics.print(L.SCRIPTEDITOR .. " - " .. scriptname, 8, 8)
		love.graphics.print(L.FILE .. "   " .. L.EDIT .. "   " .. L.INSERT .. "   " .. L.HELP, 8, 24)
		-- Script display starts at 8, 48
		--for k,v in pairs(scriptlines) do
		
		-- This can roll over, prevent that!
		textlinestogo = 0
		
		-- Make sure to display all lines but if we put the cursor further, then do display line numbers
		for k = 1, math.max(table.maxn(scriptlines), editingline) do
			v = anythingbutnil(scriptlines[k])
			
			if editingline == k then
				love.graphics.setColor(255,255,255,255)
			else
				love.graphics.setColor(128,128,128,255)
			end
			love.graphics.print(fixdig(k, 3), 8, 48+(8*k))
			
			
			textq = syntaxhl(v, 48, 48+(8*k), textlinestogo > 0, editingline == k)
			
			if textq ~= nil then
				textlinestogo = textq
			elseif textlinestogo > 0 then
				textlinestogo = textlinestogo - 1
			end
		end
	elseif state == 4 then
		--love.graphics.print(xmldoc.documentElement.childNodes[1].getAttribute("name"), 10, 10)
		love.graphics.print(metadata.Creator, 10, 10)
		love.graphics.print("That should say Unknown.", 10, 20)
	elseif state == 5 then
		love.graphics.print("Userprofile: " .. userprofile, 8, 8)
		for k,v in pairs(files) do
			love.graphics.print(v, 8, 16+8*k)
			
			lastk = k
		end
		
		--love.graphics.print(userprofile .. "\\Documentus" .. (directory_exists("C:\\Users\\" .. username .. "\\Documentus") and " EXISTS" or " DOES NOT EXIST"), 8, 16+8*lastk+16)
		--love.graphics.print(userprofile .. "\\My Documentus" .. (directory_exists("C:\\Users\\" .. username .. "\\My Documentus") and " EXISTS" or " DOES NOT EXIST"), 8, 16+8*lastk+24)

		--love.graphics.print(userprofile .. "\\Documents" .. anythingbutnil(userdatathing(directory_exists(userprofile .. "\\Documents"))), 8, 16+8*lastk+16)
		--love.graphics.print(userprofile .. "\\My Documents" .. anythingbutnil(userdatathing(directory_exists(userprofile .. "\\My Documents"))), 8, 16+8*lastk+24)
		
		love.graphics.print("Levels folder error: " .. lerror .. " (0 means no error)", 8, 16+8*lastk+16)
		love.graphics.print("Levels folder: " .. anythingbutnil(levelsfolder), 8, 16+8*lastk+24)
	elseif state == 6 then
		love.graphics.print("Levels", 8, 8)
		
		if lerror ~= 0 then
			love.graphics.printf("Could not get contents of levels folder. Please check if " .. levelsfolder .. " exists and try again.", 8, 24, love.graphics.getWidth()-16, "left")
		else
			for k,v in pairs(files) do
				love.graphics.print(v, 8, 16+8*k)
				
				lastk = k
			end
		end
		
		love.graphics.print("Load this level: " .. input .. __, 10, love.graphics.getHeight()-18)
		startinputonce()

		--love.graphics.print("Levels folder error: " .. lerror .. " (0 means no error)", 8, 16+8*lastk+16)
		--love.graphics.print("Levels folder: " .. anythingbutnil(levelsfolder), 8, 16+8*lastk+24)
	elseif state == 7 then
		for y = 0, 7 do
			for x = 0, 23 do
				love.graphics.draw(tilesets["sprites.png"]["img"], tilesets["sprites.png"]["tiles"][(y*24)+x], x*32, y*32)
				
				if mouseon(x*32, y*32, 32, 32) then
					love.graphics.print((y*24)+x, love.graphics.getWidth()-24, love.graphics.getHeight()-8)
				end
			end
		end
	elseif state == 8 then
		love.graphics.print("Enter name to save as: " .. input .. __, 10, 10)
		startinputonce()
		if savedsuccess ~= nil then
			if savedsuccess == true then
				love.graphics.print("SUCCESS!", 10, 50)
			else
				love.graphics.print("NOT SUCCESSFUL! Error: " .. anythingbutnil(savederror), 10, 50)
			end
		end
	elseif state == 9 then
		love.graphics.print(youdidanswer .. "\nRight click menu return: " .. anythingbutnil(RCMreturn), 10, 10)
	else
		fatalerror(langkeys(L.UNKNOWNSTATE, {state, oldstate}))
	end
	
	rightclickmenu.draw()
	
	dialog.draw()
	love.graphics.setColor(255,255,255,255)
	
	-- Show debug code if wanted
	if allowdebug then
		
	end
end

function love.update()
	if state ~= -1 then
		if allowdebug then
			love.window.setTitle("*Untitled - Ved  [FPS: " .. love.timer.getFPS() .. "] - State: " .. state .. " - " .. love.graphics.getWidth() .. "x" .. love.graphics.getHeight() .. " Mouse: " .. love.mouse.getX() .. " " .. love.mouse.getY())
		else
			love.window.setTitle("*Untitled - Ved")
		end
	end
	
	if RCMactive or (DIAwindowani ~= 16) then
		nodialog = false
	elseif not love.mouse.isDown("l") then
		nodialog = true
	end
	
	-- Right click menu
	if RCMreturn ~= nil and RCMreturn ~= "" then
		if RCMid:sub(1, 4) == "ent_" then
			-- Something to do with an entity.
			entdetails = explode("_", RCMid)
			if RCMreturn == L.DELETE then
				-- Delete this entity, whatever it is.
				entitydata[tonumber(entdetails[3])] = nil
			elseif RCMreturn == L.PROPERTIES then
				-- Edit properties of this entity, whatever it is.
				thisentity = entitydata[tonumber(entdetails[3])]
				startmultiinput({thisentity.x, thisentity.y, thisentity.t, thisentity.p1, thisentity.p2, thisentity.p3, thisentity.p4, thisentity.p5, thisentity.p6, thisentity.data})
				dialog.new("You can change the raw property values of this entity here.\n\nx\ny\nt\np1\np2\np3\np4\np5\np6\ndata", "", 1, 5, 2)
			elseif tonumber(entdetails[2]) == 1 then
				-- Enemy
				dialog.new(RCMid .. " " .. RCMreturn .. " not supported yet.", "", 1, 1, 0)
			elseif tonumber(entdetails[2]) == 2 then
				-- Platform, moving/conveyor
				dialog.new(RCMid .. " " .. RCMreturn .. " not supported yet.", "", 1, 1, 0)
			elseif tonumber(entdetails[2]) == 10 then
				-- Checkpoint
				dialog.new(RCMid .. " " .. RCMreturn .. " not supported yet.", "", 1, 1, 0)
			elseif tonumber(entdetails[2]) == 11 then
				-- Gravity line
				dialog.new(RCMid .. " " .. RCMreturn .. " not supported yet.", "", 1, 1, 0)
			elseif tonumber(entdetails[2]) == 13 then
				-- Warp token
				if RCMreturn == L.GOTODESTINATION then
					roomx = math.floor(entitydata[tonumber(entdetails[3])].p1 / 40)
					roomy = math.floor(entitydata[tonumber(entdetails[3])].p2 / 30)
					love.mouse.setPosition(64 + (entitydata[tonumber(entdetails[3])].p1 - (roomx*40))*16 + 8, (entitydata[tonumber(entdetails[3])].p2 - (roomy*30))*16 + 8)
					cons("Destination token is at " .. entitydata[tonumber(entdetails[3])].p1 .. " " .. entitydata[tonumber(entdetails[3])].p2 .. "... So at " .. entitydata[tonumber(entdetails[3])].p1 - (roomx*40) .. " " .. entitydata[tonumber(entdetails[3])].p2 - (roomy*30) .. " in room " .. roomx .. " " .. roomy)
				elseif RCMreturn == L.GOTOENTRANCE then
					roomx = math.floor(entitydata[tonumber(entdetails[3])].x / 40)
					roomy = math.floor(entitydata[tonumber(entdetails[3])].y / 30)
					love.mouse.setPosition(64 + (entitydata[tonumber(entdetails[3])].x - (roomx*40))*16 + 8, (entitydata[tonumber(entdetails[3])].y - (roomy*30))*16 + 8)
					cons("Entrance token is at " .. entitydata[tonumber(entdetails[3])].x .. " " .. entitydata[tonumber(entdetails[3])].y .. "... So at " .. entitydata[tonumber(entdetails[3])].x - (roomx*40) .. " " .. entitydata[tonumber(entdetails[3])].y - (roomy*30) .. " in room " .. roomx .. " " .. roomy)
				else
					dialog.new(RCMid .. " " .. RCMreturn .. " not supported yet.", "", 1, 1, 0)
				end
			elseif tonumber(entdetails[2]) == 15 then
				-- Rescuable crewmate
				dialog.new(RCMid .. " " .. RCMreturn .. " not supported yet.", "", 1, 1, 0)
			elseif tonumber(entdetails[2]) == 16 then
				-- Start point
				dialog.new(RCMid .. " " .. RCMreturn .. " not supported yet.", "", 1, 1, 0)
			elseif tonumber(entdetails[2]) == 17 then
				-- Roomtext
				dialog.new(RCMid .. " " .. RCMreturn .. " not supported yet.", "", 1, 1, 0)
			elseif tonumber(entdetails[2]) == 18 then
				-- Terminal
				if RCMreturn == L.EDITSCRIPT then
					scriptname = entitydata[tonumber(entdetails[3])].data
					scriptlines = scripts[entitydata[tonumber(entdetails[3])].data]
					tostate(3)
				else
					dialog.new(RCMid .. " " .. RCMreturn .. " not supported yet.", "", 1, 1, 0)
				end
			elseif tonumber(entdetails[2]) == 19 then
				-- Script box
				if RCMreturn == L.EDITSCRIPT then
					scriptname = entitydata[tonumber(entdetails[3])].data
					scriptlines = scripts[entitydata[tonumber(entdetails[3])].data]
					tostate(3)
				else
					dialog.new(RCMid .. " " .. RCMreturn .. " not supported yet.", "", 1, 1, 0)
				end
			end
		end
		
		RCMreturn = ""
	end
	
	dialog.update()
end

function love.textinput(char)
	if takinginput then
		input = input .. char
		
		if state == 3 then
			scriptlines[editingline] = input
		end
	elseif currentmultiinput ~= 0 then
		multiinput[currentmultiinput] = multiinput[currentmultiinput] .. char
	end
end

function love.keypressed(key)
	if takinginput then
		if key == "backspace" then
			input = backspace(input)
			
			if state == 3 then
				scriptlines[editingline] = input
			end
		end
	elseif currentmultiinput ~= 0 then
		if key == "backspace" then
			multiinput[currentmultiinput] = backspace(multiinput[currentmultiinput])
		end
	end
	
	if (DIAwindowani ~= 16) and (DIAcanclose == 1) and (key == "return") then
		dialog.push()
		DIAreturn = 1
	elseif (DIAwindowani ~= 16) and (DIAcanclose == 2) and (key == "return") then
		dialog.push()
		DIAreturn = 1
		DIAquitting = 1
	elseif (DIAwindowani ~= 16) and (DIAcanclose == 4) and (key == "return") then
		dialog.push()
		DIAreturn = 2
	elseif (DIAwindowani ~= 16) and (DIAcanclose == 5) and (key == "return") then
		dialog.push()
		DIAreturn = 3
	elseif (state == 0) and (key == "return") then
		stopinput()
		tostate(input)
	elseif (state == 1) and (key == "z") then -- temporary key, might change
		zoomscale = 2
	elseif (state == 1) and (key == "s") then -- again, but cycle tileset
		--metadata.tileset = cycle(metadata.tileset, 2)
		if usedtilesets[levelmetadata[(roomy)*20 + (roomx+1)].tileset] == 1 then
			levelmetadata[(roomy)*20 + (roomx+1)].tileset = 1
		else
			levelmetadata[(roomy)*20 + (roomx+1)].tileset = 0
		end
	elseif (state == 1) and (key == "t") then -- ^
		tilespicker = not tilespicker
	elseif (state == 1) and (key == "right") then
		-->
		if roomx+1 >= metadata.mapwidth then
			roomx = 0
		else
			roomx = roomx + 1
		end
	elseif (state == 1) and (key == "left") then
		--<
		if roomx+1 <= 1 then
			roomx = metadata.mapwidth-1
		else
			roomx = roomx - 1
		end
	elseif (state == 1) and (key == "down") then
		--v
		if roomy+1 >= metadata.mapheight then
			roomy = 0
		else
			roomy = roomy + 1
		end
	elseif (state == 1) and (key == "up") then
		--^
		if roomy+1 <= 1 then
			roomy = metadata.mapheight-1
		else
			roomy = roomy - 1
		end
	elseif allowdebug and (state == 1) and (key == "\\") then
		cons("*** TILESET COLOR CREATOR STARTED FOR TILESET " .. metadata.tileset .. " ***")
		cons("First select the wall tiles")
		
		tilescreator = true
		cb = {}
		ca = {}
		cs = {}
		creatorstep = 1
		creatorsubstep = 1
		
		selectedtileset = "creator"
		selectedcolor = "creator"
		
		tilesetblocks.creator.tileimg = metadata.tileset
		
		tilespicker = true
		selectedtool = 1
		
		mousepressed = false
	elseif allowdebug and (state == 1) and (key == "'") then
		-- Just display all tilesets and colors in the console.
		for k,v in pairs(tilesetblocks) do
			cons("==== " .. k .. " ====")
			for k2,v2 in pairs(v.colors) do
				cons("-> " .. k2)
			end
		end
	elseif (state == 3) and (key == "up") and (editingline > 1) then
		editingline = editingline - 1
		input = anythingbutnil(scriptlines[editingline])
	elseif (state == 3) and ((key == "down") or (key == "return")) then
		editingline = editingline + 1
		input = anythingbutnil(scriptlines[editingline])
	elseif (state == 6) and (key == "return") then
		stopinput()
		success, metadata, roomdata, entitydata, levelmetadata, scripts = loadlevel(input .. ".vvvvvv")
		tostate(1)
	elseif (state == 8) and (key == "return") then
		stopinput()
		savedsuccess, savederror = savelevel(input .. ".vvvvvv", metadata, roomdata, entitydata, levelmetadata, scripts)
	elseif allowdebug and (key == "f12") then
		tostate(0)
	elseif allowdebug and (key == "f9") then
		-- Do remove this, actually
		dialog.new("You are about to place more than twenty trinkets in this level. Doing so might cause glitches within the game.\n\nAre you sure you want to continue?", "", 1, 3, 0)
	end
end

function love.mousepressed(x, y, button)
	if (DIAwindowani ~= 16) and (DIAbar == 1) and (button == "l") and (x >= DIAx) and (x <= DIAx+DIAwidth) and (y >= DIAy-17) and (y <= DIAy) then
		DIAmovingwindow = 1
		DIAmovedfromwx = DIAx
		DIAmovedfromwy = DIAy
		DIAmovedfrommx = x
		DIAmovedfrommy = y
	end
	
	if DIAwindowani ~= 16 then
		if (DIAcanclose == 1) and mousein(DIAx+DIAwidth-51, DIAy+DIAwindowani+DIAheight-26, DIAx+DIAwidth-1, DIAy+DIAwindowani+DIAheight-1) then
			dialog.push()
			DIAreturn = 1
		elseif (DIAcanclose == 2) and mousein(DIAx+DIAwidth-51, DIAy+DIAwindowani+DIAheight-26, DIAx+DIAwidth-1, DIAy+DIAwindowani+DIAheight-1) then
			dialog.push()
			DIAreturn = 1
			DIAquitting = 1
		elseif (DIAcanclose == 3 or DIAcanclose == 4) and mousein(DIAx+DIAwidth-51, DIAy+DIAwindowani+DIAheight-26, DIAx+DIAwidth-1, DIAy+DIAwindowani+DIAheight-1) then
			dialog.push()
			DIAreturn = 1
		elseif (DIAcanclose == 5) and mousein(DIAx+DIAwidth-51, DIAy+DIAwindowani+DIAheight-26, DIAx+DIAwidth-1, DIAy+DIAwindowani+DIAheight-1) then
			DIAreturn = 1
		elseif (DIAcanclose == 3 or DIAcanclose == 4 or DIAcanclose == 5) and mousein(DIAx+DIAwidth-106, DIAy+DIAwindowani+DIAheight-26, DIAx+DIAwidth-56, DIAy+DIAwindowani+DIAheight-1) then
			dialog.push()
			DIAreturn = 2
		elseif (DIAcanclose == 5) and mousein(DIAx+DIAwidth-161, DIAy+DIAwindowani+DIAheight-26, DIAx+DIAwidth-106, DIAy+DIAwindowani+DIAheight-1) then
			dialog.push()
			DIAreturn = 3
		end
		
		return
	end
	
	if state == 1 then
		if x < 64 then
			if button == "wu" then
				lefttoolscroll = lefttoolscroll + 16
				lefttoolscrollbounds()
			elseif button == "wd" then
				lefttoolscroll = lefttoolscroll - 16
				lefttoolscrollbounds()
			end
		elseif mouseon(love.graphics.getWidth()-(7*16)-1, love.graphics.getHeight()-16-32-2-12, (6*16), 8+4) then -- tileset picker
			if dropdown == 1 then
				dropdown = 0
			else
				--dropdown = 1
				rightclickmenu.create({"#1", "Space St.", "#2", "Outside", "Lab", "Warp Zone", "Ship"}, "tileset")
			end
		elseif mouseon(love.graphics.getWidth()-(7*16)-1, love.graphics.getHeight()-16-16-2-4, (6*16), 8+4) then -- show all tiles
			tilespicker = not tilespicker
		end
	elseif state == 9 and button == "r" then
		rightclickmenu.create({"Delete", "Edit script", "Rename"}, 1)
	end
end

function love.mousereleased(x, y, button)
	mousepressed = false
	
	if DIAmovingwindow == 1 then
		DIAx = DIAmovedfromwx + (x-DIAmovedfrommx)
		DIAy = DIAmovedfromwy + (y-DIAmovedfrommy)
		DIAmovingwindow = 0
	end
end