-- First some things that could as well have gone in const.lua but are here for easy .
metadataitems =
	{
	"Creator", -- name
	"Title",
	"Created", -- 2?
	"Modified", -- empty?
	"Modifiers", -- 2?
	"Desc1", "Desc2", "Desc3",
	"website"
	}

function loadlevel(path)
	-- Returns (bool)success, (array)metadata, (array)contents, (array)entities, (array)levelmetadata, (array)scripts
	-- Map size and music is gonna move in with the metadata here.
	-- Contents is basically the raw array exploded
	-- Entities bestaat uit arrays (entity contents are array item data)
	-- Of course levelmetadata is all 400 rooms and also consists of arrays (roomname is array item roomname)
	-- Scripts are pre-exploded and scripts[scriptname] = (array)contents
	-- If loading isn't successful, metadata will be an error string.
	
	-- Ok we have the file at path...
	--[[
	if not love.filesystem.isFile(path) then
		-- Doesn't exist.
		cons("File " .. path .. " doesn't exist!")
		return false
	end
	
	-- Now load the level contents!
	contents = love.filesystem.read(path)
	]]
	
	success, contents = readlevelfile(levelsfolder .. dirsep .. path)
	
	if not success then
		return false, contents
	end
	
	x = {}
	
	-- First do the metadata.
	cons("Loading metadata...")
	x.metadata = contents:match("<MetaData>(.*)</MetaData>")
	thismetadata = {}
	for _,v in pairs(metadataitems) do
		thismetadata[v] = unxmlspecialchars(x.metadata:match("<" .. v .. ">(.*)</" .. v .. ">"))
	end
	
	-- But we'll have room size and music also move in with the metadata. Maybe change the regex so it'll match numbers only (not for music though because starting with an internal song number is possible), for now that's not necessary.
	cons("Loading room size and music...")
	thismetadata.mapwidth = tonumber(contents:match("<mapwidth>(.*)</mapwidth>"))
	thismetadata.mapheight = tonumber(contents:match("<mapheight>(.*)</mapheight>"))
	thismetadata.levmusic = tonumber(contents:match("<levmusic>(.*)</levmusic>"))
	
	-- TEMPORARY
	-- thismetadata.tileset = 1
	
	-- Now, the contents!
	cons("Loading all the contents...")
	--x.alltiles = explode(",", contents:match("<contents>(.*)</contents>"))
	
	-- Ok, explode() is far too inefficient, what else have we got?
	x.alltiles = {}
	for num in contents:match("<contents>(.*)</contents>"):gmatch("(%d+),") do
		--print(num)
		table.insert(x.alltiles, num)
	end
	
	cons("Contents split")
	
	-- Ok we need to correctly set all rooms... Rooms have 1200 tiles
	theserooms = {}
	for yk = 0, thismetadata.mapheight-1 do
		--print("Y: " .. yk)
		theserooms[yk] = {}
		for xk = 0, thismetadata.mapwidth-1 do
			theserooms[yk][xk] = {}
			for yt = 0, 29 do
				for xt = 0, 39 do
					theserooms[yk][xk][(40*yt) + xt + 1] = tonumber(anythingbutnil0(x.alltiles[(yk*1200*thismetadata.mapwidth) + (xk*40) + (yt*thismetadata.mapwidth*40) + xt + 1]))
					--cons("Tile loaded: " .. (yk*1200*thismetadata.mapwidth) + (xk*40) + (yt*thismetadata.mapwidth*40) + xt + 1 .. " (" .. xk .. " " .. yk .. " " .. xt .. " " .. yt .. "), " .. yk .. "*1200*" .. thismetadata.mapwidth .. " + " .. xk .. "*40 + " .. yt .. "*" .. thismetadata.mapwidth .. "*40 + " .. xt .. " + 1")
				end
			end
		end
	end
	
	-- Entities.
	cons("Loading entities...")
	if contents:find("<edEntities />") == nil then
		-- We have entities!
		x.entities = contents:match("<edEntities>(.*)</edEntities>")
		
		-- Get all entities
		allentities = {}
		entityid = 0
		for entity in x.entities:gmatch([[<edentity (.-)
            </edentity>]]) do
			entityid = entityid + 1
			allentities[entityid] = {}
			
			-- We now got x="x" ... p6="x">Data... Attributes to the left of the >, data to the right of it.
			metaparts = explode(">", entity)
			
			-- Explode more
			attributes = explode(" ", metaparts[1])
			
			for k,v in pairs(attributes) do
				-- Explode yet even more
				keyvalue = explode("=", v)
				
				-- Leave out the quotes and convert it to number
				allentities[entityid][keyvalue[1]] = tonumber(keyvalue[2]:sub(2, -2))
			end
			
			-- Now we only need the data...
			allentities[entityid].data = unxmlspecialchars(metaparts[2])
		end
	else
		allentities = {}
	end
	
	-- Level meta data
	cons("Loading room metadata...")
	x.levelmetadata = contents:match("<levelMetaData>(.*)</levelMetaData>")
	
	-- Get every room now.
	theselevelmetadata = {}
	croom = 0
	for room in x.levelmetadata:gmatch("<edLevelClass (.-)</edLevelClass>") do
		croom = croom + 1
		theselevelmetadata[croom] = {}
		
		-- We now got tileset="x" ... warpdir="x">Roomname... Attributes to the left of the >, roomname to the right of it.
		metaparts = explode(">", room)
		
		-- Explode more
		attributes = explode(" ", metaparts[1])
		
		for k,v in pairs(attributes) do
			-- Explode yet even more
			keyvalue = explode("=", v)
			
			-- Leave out the quotes and convert it to number
			theselevelmetadata[croom][keyvalue[1]] = tonumber(keyvalue[2]:sub(2, -2))
			
			--cons("Room: " .. croom .. " Key: " .. keyvalue[1] .. " Value: " .. tonumber(keyvalue[2]:sub(2, -2)))
		end
		
		-- Now we only need the room name...
		theselevelmetadata[croom].roomname = unxmlspecialchars(metaparts[2])
	end
	
	-- Scripts
	cons("Loading scripts...")
	x.allscripts = explode("|", unxmlspecialchars(contents:match("<script>(.*)</script>")))
	
	allscripts = {}
	
	if #x.allscripts > 1 then
		-- We don't want a crash now do we?
		currentscript = ""; sline = 1
		
		for k,v in pairs(x.allscripts) do
			if v:sub(-1, -1) == ":" then
				-- This is a script name!
				currentscript = v:sub(1, -2)
				allscripts[currentscript] = {}
				sline = 1
			else
				-- This is just a line.
				allscripts[currentscript][sline] = v
				sline = sline + 1
			end
		end
	end
	
	cons("Done loading!")
	
	-- No longer x.alltiles
	return true, thismetadata, theserooms, allentities, theselevelmetadata, allscripts
end

-- Load a template that we'll need for saving...
vvvvvvxmltemplate = love.filesystem.read("template.vvvvvv")

function savelevel(path, thismetadata, theserooms, allentities, theselevelmetadata, allscripts)
	-- Assumes we've already checked whether the file already exists and whatnot, immediately saves!
	-- Returns success, (if not) error message
	if (path == nil) or (path == "") then
		return false, "You forgot to specify a path!"
	end
	
	savethis = vvvvvvxmltemplate
	
	cons("Placing metadata...")
	for k,v in pairs(metadataitems) do
		--cons("Placing metadata: " .. v .. " which should match " .. "%$" .. string.upper(v) .. "%$ and have a value of " .. xmlspecialchars(thismetadata[v]))
		
		-- OK WHY DID YOU GIVE ME SUCH A LOOONG DEBUGGING HEADACHE?
		--[[ Why is it that
				savethis = savethis:gsub("%$" .. string.upper(v) .. "%$", xmlspecialchars(thismetadata[v]))
			is NOT good, but
				newthis = xmlspecialchars(thismetadata[v])
				savethis = savethis:gsub("%$" .. string.upper(v) .. "%$", newthis)
			IS???? ]]
		newthis = xmlspecialchars(thismetadata[v])
		savethis = savethis:gsub("%$" .. string.upper(v) .. "%$", newthis)
	end
	
	-- Hold on for a second, we need the map size and music too!
	savethis = savethis:gsub("%$MAPWIDTH%$", thismetadata["mapwidth"]):gsub("%$MAPHEIGHT%$", thismetadata["mapheight"]):gsub("%$LEVMUSIC%$", thismetadata["levmusic"])
	
	-- The contents are gonna be the hardest!
	cons("Assembling contents......")
	thenewcontents = {}
	for roomy, yv in pairs(theserooms) do
		-- We now have each y.....
		cons("Y: " .. roomy)
		for line = 0, 29 do
			-- (each line)
			for roomx, xv in pairs(yv) do
				-- .....And each x for each line
				--for tilex = 0, 39 do
					-- Maybe there's a notation for taking a specific part of an array to implode? I think it would be a whole lot better than concatenating every single tile like this
					--thenewcontents = thenewcontents .. theserooms[roomy][roomx][(line*40)+tilex+1] .. ","
				--end
				-- Heeey
				table.insert(thenewcontents, table.concat({unpack(theserooms[roomy][roomx], (line*40)+1, (line*40)+40)}, ","))
			end
		end
	end
	savethis = savethis:gsub("%$CONTENTS%$", table.concat(thenewcontents, ",") .. ",")
	
	-- Slightly cleaning up
	thenewcontents = nil
	
	-- Now do all entities, if we have any!
	cons("Saving entities...")
	if #allentities > 0 then
		-- We do!
		thenewentities = "        <edEntities>\n"
		for k,v in pairs(allentities) do
			thenewentities = thenewentities .. "            <edentity x=\"" .. v.x .. "\" y=\"" .. v.y .. "\" t=\"" .. v.t .. "\" p1=\"" .. v.p1 .. "\" p2=\"" .. v.p2 .. "\" p3=\"" .. v.p3 .. "\" p4=\"" .. v.p4 .. "\" p5=\"" .. v.p5 .. "\" p6=\"" .. v.p6 .. "\">" .. xmlspecialchars(v.data) .. "\n            </edentity>\n"
		end
		savethis = savethis:gsub("%$EDENTITIES%$", thenewentities .. "        </edEntities>")
	else
		-- We don't!
		savethis = savethis:gsub("%$EDENTITIES%$", "        <edEntities />")
	end
	
	cons("Saving room metadata...")
	-- Now all room metadata, aka levelclass
	alllevelmetadata = ""
	for k,v in pairs(theselevelmetadata) do
		alllevelmetadata = alllevelmetadata .. "            <edLevelClass tileset=\"" .. v.tileset .. "\" tilecol=\"" .. v.tilecol .. "\" platx1=\"" .. v.platx1 .. "\" platy1=\"" .. v.platy1 .. "\" platx2=\"" .. v.platx2 .. "\" platy2=\"" .. v.platy2 .. "\" platv=\"" .. v.platv .. "\" enemyx1=\"" .. v.enemyx1 .. "\" enemyy1=\"" .. v.enemyy1 .. "\" enemyx2=\"" .. v.enemyx2 .. "\" enemyy2=\"" .. v.enemyy2 .. "\" enemytype=\"" .. v.enemytype .. "\" directmode=\"" .. v.directmode .. "\" warpdir=\"" .. v.warpdir .. "\">" .. xmlspecialchars(v.roomname) .. "</edLevelClass>\n"
	end
	
	savethis = savethis:gsub("%$EDLEVELCLASSES%$", alllevelmetadata)
	
	-- Now all the scripts!
	cons("Assembling scripts...")
	allallscripts = ""
	for k,v in pairs(allscripts) do
		allallscripts = allallscripts .. xmlspecialchars(k) .. ":|" .. xmlspecialchars(implode("|", v)) .. "|"
	end
	
	savethis = savethis:gsub("%$SCRIPT%$", allallscripts)
	
	-- Alright, let's save!
	cons("Saving file...")
	success, iferrmsg = writelevelfile(levelsfolder .. dirsep .. path, savethis)
	
	return success, iferrmsg
end

function xmlspecialchars(text)
	return text:gsub("&", "&amp;"):gsub("\"", "&quot;"):gsub("'", "&apos;"):gsub("<", "&lt;"):gsub(">", "&gt;")
end

function unxmlspecialchars(text)
	return text:gsub("&gt;", ">"):gsub("&lt;", "<"):gsub("&apos;", "'"):gsub("&quot;", "\""):gsub("&amp;", "&")
end